--[[
Copyright © 2014 Stefan Stryjecki

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

]]--

local function checkSide (point, vect1, vect2)
	-- Returns a positive number if point is on the right of the line going through points vect1 and vect2,
	-- 0 if it lies on the line, and a negative number if it is on the left.
	return (vect2.x - vect1.x) * (point.y - vect1.y) - (vect2.y - vect1.y) * (point.x - vect1.x)
end

local function isInsidePolygon (point, polygon)
	-- Polygon is a list of points.
	local side = checkSide(point, polygon[#polygon], polygon[1])
	if side == 0 then 
		return "on the line" -- Returns a specific result which still is considered true by Lua.
	end
	local left = side < 0
	for i=2, #polygon do
		local left2 = checkSide(point, polygon[i-1], polygon[i]) < 0
		if left ~= left2 then
			return false
		end
	end
	return true
end

local function getBoundingBox (polygon)
	if not polygon.AABB then 
		local minx = polygon[1].x
		local miny = polygon[1].y
		local maxx = polygon[1].x
		local maxy = polygon[1].y
		for i=2, #polygon do
			if polygon[i].x < minx then
				minx = polygon[i].x
			elseif polygon[i].x > maxx then
				maxx = polygon[i].x
			end
			if polygon[i].y < miny then
				miny = polygon[i].y
			elseif polygon[i].y > maxy then
				maxy = polygon[i].y
			end
		end
		polygon.AABB = {{x=minx, y=miny}, {x=maxx, y=maxy}}
	end
	return polygon.AABB[1].x, polygon.AABB[1].y, polygon.AABB[2].x, polygon.AABB[2].y
end

local function isInsideAABB (point, polygon)
	local minx, miny, maxx, maxy = getBoundingBox(polygon)
	if point.x < minx or point.x > maxx or point.y < miny or point.x > maxx then
		return false
	else
		return true
	end
end

local function findContainingPolygon(x, y, list, key)
	local point = {x=x, y=y}
	for i=1, #list do
		local polygon
		if key then
			polygon = list[i][key]
		else
			polygon = list[i]
		end
		if (#polygon >= 3) and isInsideAABB(point, polygon) and isInsidePolygon(point, polygon) then
			return polygon, i
		end
	end
end

local function drawPolygon (vertices)
		local poly = {}
		if type(vertices) ~= "table" then
			return
		end
		for _, v in ipairs(vertices) do
			table.insert(poly, v.x)
			table.insert(poly, v.y)
		end
		if #poly >= 6 then
			love.graphics.polygon("fill", unpack(poly))
		end
end

local function printPoint (point)
	print('x=' .. point.x .. ', y=' .. point.y)
end

local function printPolygon (polygon)
	print('Printing vertices of a polygon with ' .. #polygon .. ' sides:')
	for i=1, #polygon do
		io.write('    ')
		printPoint(polygon[i])
	end
end

return {findContainingPolygon=findContainingPolygon;
		drawPolygon = drawPolygon;
		printPoint = printPoint;
		printPolygon = printPolygon}