--[[
Copyright © 2014 Stefan Stryjecki

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

]]--

Class = require "classloader"
local voronoi = require "voronoi"
local polygon = require "polygon"

--- INTERNAL FUNCTIONS ---

local function moveByAngle(vector, angle)
	local startAngle = math.atan2(vector.y, vector.x)
	local length = (vector.x^2 + vector.y^2)^0.5
	local x = length * math.cos(startAngle - angle)
	local y = length * math.sin(startAngle - angle)
	return {x=x; y=y}
end

-- Shorten p1 by m and p2 by such n that proportions are preserved.
local function slit(p1,p2, m) 
	local n = m*p2/p1 
	return p1-m, p2-n
end

local function shortenVector(vector, amount)
	amount = math.abs(amount)
	if vector.x == 0 and vector.y == 0 then
		return vector
	elseif vector.x == 0 then 
		if amount >= math.abs(vector.y) then
			return {x = 0; y = 0}
		else
			return {x = 0; y = vector.y - (vector.y < 0 and -1 or 1) * amount}
		end
	elseif vector.y == 0 then
		if amount >= math.abs(vector.x) then
			return {x = 0; y = 0}
		else
			return {x = vector.x - (vector.x < 0 and -1 or 1) * amount; y = 0}
		end
	else
		local p1, p2
		if math.abs(vector.x) > math.abs(vector.y) then
			p1 = vector.x
			p2 = vector.y
			xIsGreater = true
		else
			p1 = vector.y
			p2 = vector.x
			xIsGreater = false
		end
		if amount >= math.abs(p1) then
			return {x = 0; y = 0}
		end
		local newP1, newP2 = slit(p1, p2, (p1 < 0 and -1 or 1) * amount)
		if xIsGreater then
			return {x=newP1; y=newP2}
		else
			return {y=newP1; x=newP2}
		end
	end
end


--- CLASS DEFINITIONS ---

Class{
	name = "GridTile";
	obligatory = {"map"; "x"; "y"};
	readonly = {"map"; "x"; "y"; "midX"; "midY"; "height"; "width"; "latitude"; "sites"; 
				"oceanCoefficient"};
	initialize = function (self)
		self.__.sites = {}
		self.__.midX = (0.5 + self.x) * self.width
		self.__.midY = (0.5 + self.y) * self.width
		self.__.latitude = self.map.startLatitude - (self.midY * self.map.kmPerPixel) / self.map.kmPerLatitude
		self.__.absHeight = 0
		self.__.oceanCoefficient = 1
		self.__.groundTemperature = self.baseTemperature
		self.__.temperature = self.baseTemperature
		self.__.newTemperature = 0
		self.__.dTemperature = 0
		self.__.wind = {x=0, y=0}
	end;
	methods = {
		addSite = function(self, site)
			table.insert(self.__.sites, site)
			site.climateGrid = self
			local h = 0 -- height
			for _, s in pairs(self.sites) do
				h = h + s.absHeight
			end
			self.__.absHeight = h / #self.sites
			local w = 0 -- water
			for _, s in pairs(self.sites) do
				if s.height <= 0 then
					w = w + 1
				end
			end
			self.__.oceanCoefficient = (w / #self.sites) * self.map.oceanHeatCapacity + (#self.sites - w) / #self.sites
		end;
		recalculate = function(self) 
		-- This is the main "climate function" - it is called every day for every grid cell.
			-- 0. repairing data
			if (self.temperature ~= self.temperature) or self.temperature == (1/0) then--or math.abs(self.temperature - self.baseTemperature) > 20 then
				print("repairing temperature at", self.x, self.y, "as it was", self.temperature)
				self.temperature = self.baseTemperature
			end
			-- 1. heating / cooling air based on local irradiation
			self.temperature = (1 - self.map.climateNormalization) * self.temperature + self.map.climateNormalization * self.baseTemperature
			-- 2. exchanging temperature with ground
			local dT = (self.temperature - self.groundTemperature) * self.map.temperatureExchangeRate
			self.temperature = self.temperature - dT
			self.groundTemperature = self.groundTemperature + dT / self.oceanCoefficient
			-- creating wind
			local vector = {x=0, y=0}

			for x=self.x-1, self.x+1 do
				for y=self.y-1, self.y+1 do
					local othertemp
					if x < 0 or x > self.map.gridX or y < 0 or y > self.map.gridY then
						othertemp = self.temperature
					else
						othertemp = self.map.climate[x][y].temperature
					end
					local dTemp = math.max(0, othertemp - self.temperature)
					vector.x = vector.x + (self.x-x) * dTemp * self.map.windStrength / self.map.kmPerPixel
					vector.y = vector.y + (self.y-y) * dTemp * self.map.windStrength / self.map.kmPerPixel
				end
			end
			-- Adding Coriolis force taking into account the vector's length
			vector = moveByAngle(vector, self.map.coriolisForce[self.latitude] * (vector.x + vector.y)^2)
			-- temperature to be distributed by wind is proportional to the length of the wind vector
			local dT2 = (vector.x^2 + vector.y^2)^0.5 * self.map.windExchangeRate 
			self.wind = vector
			-- distribute temperature along the wind:
			-- 	TODO humidity/clouds need to be transfered from here as well to make sure that wind passing through an
			-- area also pushes the humidity forward
			local windvector = {x=vector.x; y=vector.y} -- copying the vector for "shortening"
			local targets = {}
			local total = 0
			repeat
				-- find the four closest points to the end of the wind vector
				local x = self.x + windvector.x / self.map.pixelsPerTile
				local y = self.y + windvector.y / self.map.pixelsPerTile
				for x1 = math.floor(x), math.ceil(x) do
					for y1 = math.floor(y), math.ceil(y) do
						if x1==inf or y1 == inf then print(x1, y1) os.exit() end
						if x1 >= 0 and x1 <= self.map.gridX and y1 >= 0 and y1 <= self.map.gridY then
							local distance = ((x-x1)^2 + (y-y1)^2)^0.5
							total = total + distance
							table.insert(targets, {x=x1; y=y1; distance = distance})
						end
					end
				end
				windvector = shortenVector(windvector, self.width * 2)
			until windvector.x < self.width * 2 and windvector.y < self.width * 2
			-- exchange temperature according to the percentages
			if #targets > 0 and dT2 ~= 0 then
				for _, t in ipairs(targets) do
					local target = self.map.climate[t.x][t.y]
					target.dTemperature = target.dTemperature + dT2 * t.distance / total
				end
			else
				dT2 = 0
			end
			-- create "new temperature" that will be taken as the basis for the next day
			self.newTemperature = self.temperature - dT2
		end;
	};
	getters = {
		width = function (self)
			return self.map.pixelsPerTile
		end;
		currentIrradiation = function (self)
			local latitude = self.latitude
			if not self.map.currentIrradiation[latitude] then
				local day = self.map.day % self.map.solarDays -- getting time of year - day 1 is the winter solstice
				self.map.currentIrradiation.dayModifier = self.map.currentIrradiation.dayModifier
					or (((self.map.solarDays/2)-
						math.abs(day - self.map.solarDays/2)) / 
						(self.map.solarDays/4)-1) * self.map.axialTilt  -- gets modifier from -axialTilt to +axialTilt
				local sunAngle = self.map.maxLatitude - math.abs(latitude - self.map.currentIrradiation.dayModifier)
				-- The angle must be "absolute" because the climateAmplitude can be non-integer
				self.map.currentIrradiation[latitude] = self.map.maxIrradiation	* math.abs(sunAngle / 90)^self.map.climateAmplitude	
			end
			return self.map.currentIrradiation[latitude] + self.map.dailyVariation
		end;
		baseTemperature = function (self)
			return self.currentIrradiation - self.height * self.map.CelsiusPerMeter - self.map.coolingFactor
		end;
		height = function (self)
			return math.max(0, self.absHeight - self.map.seaLevel)
		end;
	};
}

Class{
	name = "VoronoiMap";
	defaults = {
		minX = 0;
		minY = 0;
		boundingZone = 0;
		initialHeight = 0;
		seaLevel = 0;
		kmPerLatitude = 111;
--		CelsiusPerLatitude = 111/145; -- not needed because it is inferred in other calculations
		solarDays = 365.25;
		dayLength = 24;
		maxIrradiation = 70; -- this is in centigrade
		irradiationVariation = 2; -- also in centigrade - this is the maximum
		coolingFactor = 20; -- additional value substracted from daily temperatures
		climateAmplitude = 0.8; -- this should be between 0.5 and 3 - it increases temperature amplitudes
		axialTilt = 23.4;
		maxLatitude = 90;
		oceanHeatCapacity = 8; --4.2; -- Where ground's heat capacity is 1
		temperatureExchangeRate = 0.1; -- the higher it is, the faster the exchange between the ground and the air
		climateNormalization = 0.05; -- between 0 and 1 - the lower it is the longer do the changes in irradiation take to take effect
		windExchangeRate = 0.02; -- you have to be careful - if this is to high relative to the climateNormalization, the world will become a jumbled mess
		windStrength = 10; -- be careful - kmPerPixel affects windStrength - this should ALWAYS be smaller than kmPerPixel
		CelsiusPerMeter = 1/220;
		startLatitude = 80;
		kmPerPixel = 20;
		coriolisSimulationCoefficient = 8; -- in hours - 24 is the "logical" value as the simulation runs on a per-day basis but smaller values give more "well-behaved" wind patterns
		seed = os.time()
		
	};
	obligatory = {"maxX", "maxY"};
	readonly = {"minX", "minY", "maxX", "maxY", "gridX", "gridY", "boundingZone", "minHeight", "maxHeight", "amountOfPoints", "pixelsPerTile", "seed"};
	initialize = function (self)
		print(self.seed)
		math.randomseed(self.seed)
		self.day = 1 -- Day 1 is the winter solstice
		local points = {}
		if self.amountOfPoints then
			for i=1, self.amountOfPoints do
				points[i] = {
					x=math.random(self.minX, self.maxX)+math.random(), 
					y=math.random(self.minY, self.maxY)+math.random()}
			end
		elseif self.pointsList then
			points = self.pointsList
		else
			error("You need to specify the amountOfPoints or a pointsList")
		end
		self.sites, self.vertices
			= voronoi(points, self.boundingZone, self.minX, self.minY, self.maxX, self.maxY)
		self.__.pointsList = nil
		self.__.amountOfPoints = #self.sites
		for i=1, #self.sites do
			self.sites[i].absHeight = self.initialHeight
			self.sites[i].map = self
		end
	end;
	methods = {
		findSite = function (self, x, y)
			return polygon.findContainingPolygon(x, y, self.sites, "vertices")
		end;
		generateLandmass = function (self, size, maxchange)
			local last = self.sites[math.random(#self.sites)]
			for i=1, size do
				last.absHeight = last.absHeight + math.random(maxchange)
				last = last.neighbors[math.random(#last.neighbors)]
				if not last or last.x < self.boundingZone or last.x > self.maxX-self.boundingZone or 
					last.y < self.boundingZone or last.y > self.maxY-self.boundingZone then
					break
				end
			end
		end;
		addPseudoTectonics = function (self, threshold)
			for i=1, #self.sites do
				local distance = 0
				for _, n in pairs(self.sites[i].neighbors) do
					distance = distance + (self.sites[i].x-n.x)^2 + (self.sites[i].y-n.y)^2
				end
				distance = distance / (#self.sites[i].neighbors)
				local change = threshold^2 / distance
				self.sites[i].absHeight = self.sites[i].absHeight + change
			end
		end;
		smootheHeightMap = function (self, factor)
			factor = factor or 0
			for i=1, #self.sites do
				local target = 0
				for j=1, #self.sites[i].neighbors do
					target = target + self.sites[i].neighbors[j].absHeight
				end
				target = target / #self.sites[i].neighbors
				if (target / self.sites[i].absHeight) % 1 > factor then
					self.sites[i].absHeight = (self.sites[i].absHeight + target)/2
				end
			end
		end;
		normalizeHeights = function (self, min, max)
			local oldmin = self.minHeight
			local step = (max-min) / (self.maxHeight-oldmin)
			for i=1, #self.sites do
				self.sites[i].absHeight = (self.sites[i].absHeight - oldmin) * step + min
			end
		end;
		generateHeightMap = function (self, granularity, nLandmasses, smoothingFactor)
			for i=1, nLandmasses do
				self:generateLandmass(granularity, nLandmasses / 100)
			end
			for i=1, 7 do
				self:smootheHeightMap(smoothingFactor)
			end
			self:normalizeHeights(0, math.random(8000, 16000))
		end;
		drawSite = function (self, site)
			local poly = {}
			if type(site) == "number" then
				site = self.sites[site]
			end
			for _, v in ipairs(site.vertices) do
				table.insert(poly, v.x)
				table.insert(poly, v.y)
			end
			if #poly >= 6 then
				love.graphics.polygon("fill", unpack(poly))
			end
		end;
		getGrid = function (self, x)
			x = (type(x) == "table") and x.x or x
			return math.floor(x / self.pixelsPerTile)
		end;
		initializeClimate = function (self, pixelsPerTile)
			self.currentIrradiation = {}
			self.dailyVariation = 0
			self.__.pixelsPerTile = pixelsPerTile
			self.climate = {}
			self.__.gridX = self:getGrid(self.maxX)
			self.__.gridY = self:getGrid(self.maxY)
			for x=0, self.gridX do
				self.climate[x] = {}
				for y=0, self.gridY do
					self.climate[x][y] = GridTile{map=self; x=x; y=y}
				end
			end
			for i=1, #self.sites do 
				local x = self:getGrid(self.sites[i].x)
				local y = self:getGrid(self.sites[i].y)
				self.climate[x][y]:addSite(self.sites[i])
			end
			self.coriolisForce = {}
			for y=0, self.gridY do
				local lat = self.climate[1][y].latitude
				self.coriolisForce[lat] = 2* math.pi / (self.dayLength * 3600)  * math.sin(lat * math.pi/180) * 3600 * self.coriolisSimulationCoefficient -- *3600 - simulating effects over one hourr
			end
		end;
		moveForward = function (self)
			self.day = self.day + 1
			self.currentIrradiation = {}
			self.dailyVariation = math.random(-self.irradiationVariation, self.irradiationVariation)
			for x=0, self.gridX do
				for y=0, self.gridY do
					local c = self.climate[x][y]
					c.temperature = c.newTemperature + c.dTemperature
					c.newTemperature = 0
					c.dTemperature = 0
				end
			end
			for x=0, self.gridX do
				for y=0, self.gridY do
					self.climate[x][y]:recalculate(dailyVariation)
				end
			end
		end;
	};
	getters = {
		minHeight = function (self)
			local min = self.sites[1].absHeight
			for i=2, #self.sites do
				min = math.min(min, self.sites[i].absHeight)
			end
			return min
		end;
		maxHeight = function (self)
			local max = self.sites[1].absHeight
			for i=2, #self.sites do
				max = math.max(max, self.sites[i].absHeight)
			end
			return max
		end;
	};
}

